;
;  BICEM - Kommandozeile
;
;  Dieses Programm stellt eine Kommandozeile fr BICEM zur Verfgung.
;  (CCP - Console Command Processor im CP/M-Jargon). Das Programm
;  ldt sich vollstndig resident. Eingebaute Kommandos sind `EXIT'
;  (BICEM anhalten) und `DIR wildcard'.
;
;  Dieses Programm gilt als Teil des Kern-Emulators: nur mit CCP kann
;  man dem gestarteten Programm Parameter bergeben, und Software-
;  Implementationen der Systemdienste gehren ebenfalls hierher. Die
;  Readkey- und Readline-Dienste sollten als Z80-Code implementiert
;  werden, um die Interrupt-Semantik (und Unterbrechbarkeit im Monitor)
;  zu erhalten.
;
;  (c) copyright 1999 Stefan Reuther
;

                db      $EB,$03         ; `jmp x86entry' fr x86, `ex de,hl/inc bc' fr Z80
                jmp     Start
x86entry:       db      $B4,9           ; `ld ah,9'
                db      $BA             ; `ld dx,'
                dw      NotBICEMtxt
                db      $CD,$21,$C3     ; `int 21/ret'

Start:          ld      hl,(1)          ; Prfung, ob direkt aus BICEM
                ld      a,(hl)          ; gestartet:
                cmp     $ED             ; der jmp 0 zeigt dann auf einen
                jpnz    NotBICEM        ; BICEM-Opcode (ED FF)
                inc     hl
                ld      a,(hl)
                cmp     $FE
                jpz     InitCCP

NotBICEM:       ld      c,9
                ld      de,NotBICEMtxt
                call    5
                ret

InitCCP:        ld      c,9
                ld      de,Welcome
                call    5

                ld      sp,StackEnd

;----------------------
; Code resident machen
;----------------------
                ld      hl,(6)
                dec     hl
                ex      de,hl
                ld      hl,ResidentLast-1
                ld      bc,ResidentLast - ResidentFirst
                ; HL = letztes zu kopierendes Byte
                ; DE = letzter freier Platz
                lddr
;------------
; Relozieren
;------------
                scf
                ccf
                push    de              ; fr Einsprungpunkt merken
                ex      de,hl
                sbc     hl,de
                ex      de,hl           ; DE = ntige Verschiebung
                ld      bc,Relocations  ; Tabelle
RelocLoop:      ld      a,(bc)
                ld      l,a
                inc     bc
                ld      a,(bc)
                ld      h,a             ; HL = Patchadresse
                inc     bc
                or      l               ; 0 => Ende
                jpz     EndReloc

                add     hl,de           ; HL = Codeadresse
                push    hl
                pop     ix
                ld      a,(ix+0)        ; Addition (ix) += DE
                add     e
                ld      (ix+0),a
                ld      a,(ix+1)
                adc     d
                ld      (ix+1),a
                jmp     RelocLoop
;------------------
; Ende Relozierung
;------------------
EndReloc:       push    de              ; Verschiebung
                db      $ED,$FC         ; dem Emulator mitteilen

                pop     hl              ; HL zeigt vor Einsprungpunkt
                inc     hl
                ld      (6),hl
                inc     hl
                inc     hl
                inc     hl              ; HL = CCPLoop
                ld      ix,(1)
                ld      (ix),$C3        ; `reboot' Call
                ld      (ix+1),l
                ld      (ix+2),h
                jmp     (hl)

NotBICEMtxt:    db      'This program must be run directly under BICEM.',13,10,36
Welcome:        db      'BICEM Console Command Processor',13,10
                db      13,10
                db      '(c) copyright 1999 Stefan Reuther',13,10,36

Relocations:    dw      ResidentFirst+1
                dw      CCPloop+1
                dw      Rel1+1
                dw      Rel2+1
                dw      Rel3+1
                dw      Rel4+1
                dw      Rel5+1
                dw      Rel6+1
                dw      Rel7+1
                dw      Rel8+1
                dw      Rel9+1
                dw      Rel10+1
                dw      Rel11+1
                dw      Rel12+1
                dw      Rel13+1
                dw      Rel14+1
                dw      Rel15+1
                dw      Rel16+1
                dw      Rel17+1
                dw      Rel18+1
                dw      Rel19+1
                dw      Builtins
                dw      Rel20+1
                dw      Rel21
                dw      NoFileFound+1
                dw      Rel22+1
                dw      Rel23+1
                dw      Rel24+1
                dw      Rel25+1
                dw      0

;============================================================================
; Residenter Code
;============================================================================
ResidentFirst:  jmp     Call5
CCPLoop:        ld      sp,StackEnd
                ld      c,2
                ld      e,10
                call    5
                ld      e,13
                call    5
CCPLoop1:       ld      c,25            ; Abfrage Bezugs-LW
                call    5
                add     65
                ld      c,2             ; Zeichenausgabe
                ld      e,a
                call    5
                ld      e,58            ; ':'
                call    5
                ld      e,62            ; '>'
                call    5
  rel1:         ld      de,InputBuffer
                ld      c,10
                call    5
;---------------------
; Eingabe vorbereiten
;---------------------
                inc     de              ; DE => Lngenbyte
                ld      a,(de)
                cmp     0
                jrz     CCPLoop1        ; Keine Eingabe => Nochmal
                ld      b,a
                inc     de
                push    de              ; (sp) = Textadresse
  PrepareLoop:  ld      a,(de)
                cmp     97              ; 'a'
                jrc     NotLower
                cmp     123             ; 'z'+1
                jrnc    NotLower
                sub     32
                ld      (de),a
  NotLower:     inc     de
                djnz    PrepareLoop
                ld      a,0
                ld      (de),a
                inc     de
                ld      (de),a

;--------------------------
; Zieldaten initialisieren
;--------------------------
  Rel5:         ld      hl,ProgFCB      ; Programmnamens-FCB
                ld      (hl),0
                inc     hl
  Rel6:         ld      de,ProgFCB+2
                ld      (hl),32
                ld      bc,7
                ldir
                inc     hl
                ld      (hl),67         ; 'C'
                inc     hl
                ld      (hl),79         ; 'O'
                inc     hl
                ld      (hl),77         ; 'M'

                ld      hl,$5C          ; User-FCBs
                ld      de,$5D          ; komplett ausnullen
                ld      bc,31
                ld      (hl),0
                ldir

                ld      l,$5D           ; Dateinamen auf leer setzen
                ld      e,$5E           ; FCB 1
                ld      (hl),32
                ld      c,10
                ldir
                ld      l,$6D           ; FCB 2
                ld      e,$6E
                ld      (hl),32
                ld      c,10
                ldir

                ld      l,$80           ; Command Tail auf Leerstring setzen
                ld      (hl),0
                inc     hl
                ld      (hl),13

;----------------------
; Kommandozeile parsen
;----------------------
                pop     ix              ; ix => Zeichen
                ld      c,6             ; c = Status: 6,5 = Befehl, 4,3 = arg1, 2,1 = arg2
                                        ; even = Name, odd = ext
  Rel2:         ld      de,ProgFCB

  ;+++++++++++++++++++++
  ; Parserschleife:
  ;  de -> FCB
  ;  ix -> Zeichen
  ;  c  = Status (siehe oben)
  ;+++++++++++++++++++++
  ParseLoop:    ld      a,58            ; ':'
                cmp     (ix+1)          ; Laufwerk angegeben?
                jrnz    NoDriveLetter
                ld      a,(ix)          ; ja. Teste Gltigkeit.
                sub     65              ; => 0 = A, ...
                jrc     InvalidDrive
                inc     a
                cmp     17
                jrc     SetDriveLetter
  InvalidDrive: ld      a,c             ; sind wir im Programmnamen?
                cmp     3               ; dann ist es ein fataler Fehler
                ld      a,0             ; sonst nehmen wir LW 0 (default) an
                jrnz    SetDriveLetter
                ld      c,9
  Rel3:         ld      de,InvalidDriveText
                call    5
  Rel4:         jmp     CCPLoop

  SetDriveLetter:                       ; hier: A = Lw fr FCB
                ld      (de),a          ; eintragen
                inc     ix              ; und in Kommandozeile berspringen
                inc     ix
  NoDriveLetter:
                inc     de              ; Feld im FCB berspringen
                ld      b,8             ; max 8 Zeichen Dateiname
  FilenameLoop:
                ld      a,(ix)
  Rel7:         call    Classify
                jr      EoFilename
                jr      Extension
                jr      FillQues
                ld      (de),a          ; Normales Zeichen:
                inc     de              ; in FCB eintragen
                inc     ix
                djnz    FilenameLoop    ; und weiter
                jr      SkipTillDot

  FillQues:     ld      a,63            ; '*': mit '?' auffllen
                ld      (de),a
                inc     de
                djnz    FillQues
                inc     ix

  SkipTillDot:  ld      a,(ix)          ; Zuviele Zeichen angegeben => berspringen
                ld      b,0             ; keine Zeichen mehr fllen
  Rel8:         call    Classify
                jr      EoFilename
                jr      Extension
                jr      SkipNextChar
  SkipNextChar: inc     ix
                jr      SkipTillDot

  Extension:    ld      a,e             ; Extension-Kennung angegeben
                add     b               ; => berspringe diesen Part des FCB
                ld      e,a             ; DE += B
                jrnc    NoCarry
                inc     d
  NoCarry:
                inc     ix              ; Punkt berlesen
                dec     c               ; Statuswert
                bit     0,c             ; gerade: wir sind schon in der Extension
                jrz     EoFilename1
                ld      b,3             ; 3 Zeichen Extension
                jr      FilenameLoop

  EoFilename:   dec     c               ; wenn keine Extension angegeben: 2 Stati
                bit     0,c             ; berspringen
                jrnz    EoFilename
  EoFilename1:
                ld      a,0             ; sind wir am Ende?
                cmp     (ix)
                jrz     ParsingDone
                cmp     c               ; kein FCB mehr zu parsen?
                jrz     ParsingDone

  SkipWSLoop:   ld      a,(ix)          ; Whitespace berlesen
                or      a
                jrz     ParsingDone
                cmp     33
                jrnc    EOWhitespace
                inc     ix
                jr      SkipWSLoop

  EOWhitespace: ld      a,4             ; OK, nchstes "token"
                cmp     c
                jrz     FillFCB1
                ld      de,$6C          ; zweiter FCB -> kopieren
  Rel9:         jmp     ParseLoop
  FillFCB1:     push    ix              ; erst mal den Command Tail erzeugen
                ld      hl,$81
                ld      b,0
  CopyCmdTail:  ld      a,(ix)
                or      a
                jrz     EOTail
                ld      (hl),a
                inc     ix
                inc     hl
                inc     b
                jr      CopyCmdTail
  EOTail:       ld      (hl),13         ; ein Enter hintendran
                ld      a,b
                ld      ($80),a         ; und die Lnge
                pop     ix
                ld      de,$5C          ; und nun den 1. FCB
  Rel10:        jmp     ParseLoop

  ParsingDone:

;==============================================================
; Ende Parsing: Nun ausfhren
;==============================================================
  Rel11:        ld      de,ProgFCB
                push    de
                pop     ix
                ld      a,(ix+1)        ; erstes Dateinamenszeichen
                cmp     32
                jrnz    NoDriveChange
                ld      c,14
                ld      e,(ix)
                dec     e
                call    5
  Rel12:        jmp     CCPLoop1

;===========================================================
; Ist es ein builtin-Kommando?
;===========================================================
  NoDriveChange:
                ld      a,0
                cmp     (ix)
                jrnz    NoBuiltin
  Rel20:        ld      hl,Builtins
  BuiltinLoop:  ld      e,(hl)          ; Adresse des Handlers
                inc     hl
                ld      d,(hl)
                inc     hl
                ld      a,d
                or      e
                jrz     NoBuiltin
                push    de              ; Adresse liegt nun auf'm Stack
                push    ix
                ld      bc,$0808        ; c = Anzahl gleiche Zeichen
  CompareLoop:  inc     ix
                ld      a,(ix)
                cmp     (hl)
                jrnz    NoDec
                dec     c
  NoDec:        inc     hl
                djnz    CompareLoop
                pop     ix
                ld      a,c
                or      a
                rz
                pop     de
                jr      BuiltinLoop

;===========================================================
; Diskettenprogramm laden
;===========================================================
  NoBuiltin:    push    ix
                pop     de
                ld      c,15
                call    5
                cmp     $FF
                jrnz    OpenedOK
  Rel13:        ld      de,Filenotfoundtext
                ld      c,9             ; Open FCB
                call    5
  Rel19:        jmp     CCPLoop
  OpenedOK:
                ld      hl,$100         ; Start TPA
  LoadLoop:     ex      de,hl
                ld      c,26            ; Set DTA
                call    5
                ex      de,hl
                ld      c,20            ; Read Seq
                call    5
                cmp     0
                jrnz    EndOfFile
                ld      bc,$80
                add     hl,bc
  Rel14:        ld      bc,ResidentFirst-$80
                push    hl
                scf
                sbc     hl,bc
                pop     hl
                jrc     LoadLoop
                ld      c,16            ; Close FCB
                call    5
  Rel15:        ld      de,TooBigText
                ld      c,9
                call    5
  Rel16:        jmp     CCPLoop

  EndOfFile:    ld      c,16            ; Datei schlieen
                call    5
                ld      c,26            ; Setze DTA
                ld      de,$80
                call    5
  Rel25:        ld      sp,StackEnd
                ld      hl,0
                push    hl
                jmp     $100

;==============================================================
; Unterprogramm: Zeichen klassifizieren
;  ld   a,Zeichen
;  call Classify
;  jr   wennDateinamensEnde
;  jr   wennPunkt
;  jr   wennStern
; wennNormal:
; Zerstrt HL
;==============================================================
Classify:       pop     hl
                push    hl
                cmp     32
                rz
                cmp     0
                rz
                cmp     46              ; '.'
                jrz     ReturnExtension
                cmp     42              ; '*'
                jrz     ReturnFillQues
                cmp     44              ; ','
                rz
                cmp     58              ; ':'
                jrc     ReturnCharOk
                cmp     63              ; '?
                rc
  ReturnCharOk: inc     hl
                inc     hl
  ReturnFillQues: inc   hl
                inc     hl
  ReturnExtension: inc  hl
                inc     hl
                ex      (sp),hl
                ret

;==========================================================================
Call5:          push    af
                ld      a,c
                cmp     0
  Rel18:        jpz     CCPLoop
                pop     af
  Rel17:        jmp     ResidentLast

;==========================================================================
ExitEmu:        db      $ED,$FE
                                                                                                              
;==========================================================================
Dir:            ld      de,$80          ; Set DTA
                ld      c,26
                call    5
                ld      hl,$5D
                ld      a,(hl)
                cmp     32
                jrnz    NeedntInitFCB
                ld      de,$5E
                ld      (hl),63         ; '?'
                ld      bc,10
                ldir
  NeedntInitFCB:ld      de,$5C
                ld      c,17            ; Find First
                call    5
                cmp     $FF
                jrz     NoFileFound
MoreFiles:      ld      hl,$81
  Rel22:        call    PrintFilename
                ld      c,2             ; Print Char
                ld      e,9
                call    5
                ld      c,18            ; Find Next
                call    5
                cmp     $FF
                jrnz    MoreFiles
                rst     0
NoFileFound:    ld      de,NoFileText1
                ld      c,9
                call    5
                ld      hl,$5D
  Rel23:        call    PrintFilename
  Rel24:        ld      de,NoFileText2
                ld      c,9
                call    5
                rst     0

PrintFilename:  ld      b,11
                ld      c,2
 PrintLoop:     ld      e,(hl)
                call    5
                inc     hl
                ld      a,4
                cmp     b
                jrnz    DontPrintDot
                ld      e,46
                call    5
DontPrintDot:   djnz    PrintLoop
                ret

;==========================================================================
Builtins:       dw      Exitemu
                db      'EXIT    '
  Rel21:        dw      Dir
                db      'DIR     '
                db      0
;==========================================================================
InputBuffer:    db      128             ; Kapazitt
                db      0               ; Fllstand
                space   128
                db      0,0             ; Platz fr 2 Fllbytes

ProgFCB:        db      0               ; Laufwerk
 ProgName:      space   8
                db      'COM'
                space   21

Stack:          space   64              ; vanilla-SCP bietet 48 bytes
StackEnd:                               ; fr transiente Programme

InvalidDriveText: db    'Invalid drive letter in command name.',13,10,36
FilenotfoundText: db    'File not found.',13,10,36
TooBigText:       db    'Not enough memory to run program.',13,10,36
NoFileText1:      db    'No files matching `$'
NoFileText2:      db    39,46,13,10,36

ResidentLast:   ; Sprung hierhin = alter CALL 5
