;
;  BICEM - Emulator-Status ausgeben
;
;  (c) copyright 1999 Stefan Reuther
;

                db      $EB,$03
                jmp     Start
                db      $B4,9,$BA
                dw      StatusText
                db      $CD,$21,$BA
                dw      NoEmuText
                db      $CD,$21,$C3
Start:          ld      de,StatusText           ; Titelzeile, "Video RAM:"
                ld      c,9
                call    5
                ld      de,VRAMtext
                call    5

                call    StatusInfo
                push    hl                      ; "brk"
                push    de                      ; "sys"

                ;=== Video-RAM Meldung ===
                ld      de,NoneText
                jrnc    NoVideoRAM
                push    bc                      ; "VRAM.at0"
                call    PrintHexNumber
                ld      de,LinefeedText
NoVideoRAM:     ld      c,9
                call    5

                ;=== "sys" - Systemcode-Adresse ===
                ld      de,SysText
                call    5
                call    PrintHexNumber

                ;=== BDOS-Eintrittspunkt ===
                ld      c,9
                ld      de,BdosText
                call    5
                ld      hl,(6)
                push    hl                      ; merken
                push    hl                      ; als Argument
                call    PrintHexNumber

                ;=== Freier Speicher ===
                pop     hl
                ld      bc,256
                scf
                sbc     hl,bc
                push    hl
                ld      de,LeftParen
                ld      c,9
                call    5
                call    PrintDecNumber

                ;=== "brk" - hchste verfgbare Adresse ===
                ld      c,9
                ld      de,BrkText
                call    5
                call    PrintHexNumber

                ;=== Information ber Timerinterrupt ===
                ld      c,9
                ld      de,TimerText
                call    5
                call    TimerInfo
                jrnc    NoTimer
                push    de                      ; Adresse
                push    bc                      ; Frequenz

                ld      c,9
                ld      de,EveryText
                call    5
                call    PrintDecNumber          ; Frequenz ausgeben

                ld      c,9
                ld      de,JumpText
                call    5
                call    PrintHexNumber          ; Adresse ausgeben

                ld      de,EndTimerText
                jr      SkipNoTimer

NoTimer:        ld      de,NoneText
SkipNoTimer:    ld      c,9
                call    5

                ret

;---------------------------------------------------------------------------
StatusInfo:     db      $ED,$FD,131
TimerInfo:      db      $ED,$FD,130
;---------------------------------------------------------------------------

;---------------------------------------------------------------------------
; Zahl auf dem Stack dezimal ausgeben; nicht reentrant
;---------------------------------------------------------------------------
PrintDecNumber: pop     hl
                ex      (sp),hl
                ld      ix,DigitBuffer          ; als fnfstellige Zahl
                ld      bc,9999                 ; formatieren
                call    Digit
                ld      bc,999
                call    Digit
                ld      bc,99
                call    Digit
                ld      c,9
                call    Digit
                ld      (ix),l

                ld      ix,DigitBuffer          ; maximal 4 fhrende Nullen
                ld      b,4                     ; entfernen
  SkipLoop:     ld      a,(ix)
                cmp     0
                jrnz    PrintLoop
                inc     ix
                djnz    SkipLoop
  PrintLoop:    ld      a,(ix)                  ; Ausgabe
                add     $30
                rc                              ; Carry entsteht, wenn der
                ld      e,a                     ; Sentinel erreicht wird
                ld      c,2
                call    5
                inc     ix
                jr      PrintLoop

;--- Unterprogramm: *ix++ = HL/(BC+1), HL %= (BC+1) durch Subtraktion ---
Digit:          ld      a,-1
  DigitLoop:    inc     a
                scf
                sbc     hl,bc
                jrnc    DigitLoop
                adc     hl,bc
                ld      (ix),a
                inc     ix
                ret

;---------------------------------------------------------------------------
; Zahl auf dem Stack als vierstellige Hexzahl ausgeben
;---------------------------------------------------------------------------
PrintHexNumber: pop     hl
                ex      (sp),hl
                ld      bc,$402         ; Funktionsnummer, Zhler
PrintHexLoop:   ld      a,h
                srl     a
                srl     a
                srl     a
                srl     a
                add     $30
                cmp     $3A
                jrc     NoLetter
                add     7
  NoLetter:     ld      e,a
                call    5
                add     hl,hl
                add     hl,hl
                add     hl,hl
                add     hl,hl
                djnz    PrintHexLoop
                ret

DigitBuffer:    space   5
                db      $FF     ; Sentinel

StatusText:     db      'BICEM - Emulator Status',13,10,10,'$'
NoEmuText:      db      'Emulator not running.',13,10,'$'
VRamText:       db      'Video RAM        : $'
SysText:        db      'System Code      : $'
BdosText:       db      13,10
                db      'BDOS Entry Point : $'
LeftParen:      db      ' ($'
BrkText:        db      ' bytes free)',13,10
                db      'Highest Address  : $'
NoneText:       db      '(none)'
LinefeedText:   db      13,10,'$'
TimerText:      db      13,10,
                db      'Timer            : $'
EveryText:      db      'every $'
JumpText:       db      ' PC timer ticks execute a `call $'
EndTimerText:   db      39,13,10,36
