UNIT StrUtil;

INTERFACE

TYPE TShortStr=STRING[15];
     TMedStr=STRING[31];

CONST s_Leading         = 1;
      s_Trailing        = 2;
      s_Both            = 3;

{=== Zeichenumsetzung ===}
FUNCTION UpStr(S:STRING):STRING;

{=== Zahlen konvertieren ===}
FUNCTION Hex(L:LONGINT; Width:INTEGER):STRING;
FUNCTION iToA(L:LONGINT):TShortStr;
FUNCTION Eval(S:STRING; VAR L:LONGINT):BOOLEAN;

{=== Tabulatoren ===}
PROCEDURE ExpandTabs(VAR S:STRING);
PROCEDURE PackTabs(VAR S:STRING);

{=== Formatierung ===}
PROCEDURE Strip(VAR S:STRING; Where:BYTE);

{=== Stringarithmetik ===}
PROCEDURE AppendChar(VAR S:STRING; C:CHAR);
PROCEDURE SetLength(VAR S:STRING; Len:BYTE);
FUNCTION StartsWith(S:STRING; SubStr:STRING):BOOLEAN;

{=== Suchmuster ===}
FUNCTION Matches(Pattern:STRING; Str:STRING):BOOLEAN;

CONST HexD:ARRAY[0..15] OF CHAR='0123456789ABCDEF';

IMPLEMENTATION

FUNCTION IsBlank(C:CHAR):BOOLEAN;
BEGIN
  IsBlank:=(C=' ') OR (C=#0) OR (C=#9);
END;

{---------------------------------------------------------------------------}
{                                                                           }
{   U p S t r                                                               }
{ - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - }
{                                                                           }
{   Zeichenkette in Grobuchstaben                                          }
{                                                                           }
{---------------------------------------------------------------------------}
FUNCTION UpStr(S:STRING):STRING;
VAR i:INTEGER;
BEGIN
  FOR i:=1 TO Length(S) DO S[i]:=Upcase(S[i]);
  UpStr:=S;
END;

{===========================================================================}
{                                                                           }
{   H e x                                                                   }
{ - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - }
{                                                                           }
{   Wandelt Zahl in Hex um                                                  }
{   Eingabe: L: Zahl (0)                                                   }
{            Width: Mindestbreite                                           }
{                                                                           }
{===========================================================================}
FUNCTION Hex(L:LONGINT; Width:INTEGER):STRING;
VAR S:STRING;
BEGIN
  S:='';
  WHILE (L<>0) OR (Width>0) DO BEGIN
    S:=HexD[L AND 15]+S;
    L:=L SHR 4;
    Dec(Width);
  END;
  IF S='' THEN S:='0';
  Hex:=S;
END;

{---------------------------------------------------------------------------}
{                                                                           }
{   i T o A                                                                 }
{ - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - }
{                                                                           }
{   Wandelt die Zahl L in einen String um.                                  }
{                                                                           }
{---------------------------------------------------------------------------}
FUNCTION iToA(L:LONGINT):TShortStr;
VAR S:TShortStr;
BEGIN
  Str(L,S);
  iToA:=S;
END;

{---------------------------------------------------------------------------}
{                                                                           }
{   E x p a n d T a b s                                                     }
{ - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - }
{                                                                           }
{   Erweitert Tabulatoren zu Leerzeichen                                    }
{                                                                           }
{---------------------------------------------------------------------------}
PROCEDURE ExpandTabs(VAR S:STRING);
VAR i:INTEGER;
BEGIN
  i:=1;
  WHILE i<=Length(S) DO BEGIN
    IF S[i]=#9 THEN BEGIN
      Insert(' ',S,i);
      IF i AND 7=0 THEN Delete(S,i+1,1);
    END;
    Inc(i);
  END;
END;

{---------------------------------------------------------------------------}
{                                                                           }
{   P a c k T a b s                                                         }
{ - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - }
{                                                                           }
{   Fat Leerzeichen zu Tabulatoren zusammen.                               }
{                                                                           }
{---------------------------------------------------------------------------}
PROCEDURE PackTabs(VAR S:STRING);
VAR i,k:INTEGER;
BEGIN
  i:=Length(S) AND (NOT 7);
  WHILE i>0 DO BEGIN
    k:=0;
    WHILE (S[i-k]=' ') AND (k<8) DO Inc(k);
    IF k>=2 THEN BEGIN
      Delete(S,i-k+1,k);
      Insert(#9,S,i-k+1);
    END;
    Dec(i,8);
  END;
END;

{---------------------------------------------------------------------------}
{                                                                           }
{   S t r i p                                                               }
{ - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - }
{                                                                           }
{   Entfernt fhrende/angehngte Leerzeichen, der Code Where gibt an,     }
{   welche.                                                                 }
{                                                                           }
{---------------------------------------------------------------------------}
PROCEDURE Strip(VAR S:STRING; Where:BYTE);
BEGIN
  IF Where AND s_Leading<>0 THEN
   WHILE (S<>'') AND IsBlank(S[1]) DO Delete(S,1,1);
  IF Where AND s_Trailing<>0 THEN
   WHILE (S<>'') AND IsBlank(S[Length(S)]) DO Dec(S[0]);
END;

{---------------------------------------------------------------------------}
{                                                                           }
{   E v a l                                                                 }
{ - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - }
{                                                                           }
{   Wertet die bergebene Zeichenkette aus und erzeugt eine entsprechende   }
{   Longint-Zahl daraus. Wenn das mglich ist, wird TRUE zurckgegeben,     }
{   sonst FALSE. Es sind die Formen "123", "123h", "123o" und "101b" mg-   }
{   lich, nicht jedoch Vorzeichen oder "$123".                              }
{                                                                           }
{---------------------------------------------------------------------------}
FUNCTION Eval(S:STRING; VAR L:LONGINT):BOOLEAN;
VAR Base:INTEGER;
    i,dig:INTEGER;
BEGIN
  Eval:=FALSE;
  Strip(S,s_Both);
  IF (S='') OR (S[1]<'0') OR (S[1]>'9') THEN Exit;
  Base:=10;
  CASE S[Length(S)] OF
    'H','h':Base:=16;
    'B','b':Base:=2;
    'O','o':Base:=8;
  END;
  IF Base<>10 THEN Dec(S[0]);
  IF S='' THEN Exit;
  FOR i:=1 TO Length(S) DO BEGIN
    CASE S[i] OF
      'A'..'F':Dig:=Ord(S[i])-55;
      'a'..'f':Dig:=Ord(S[i])-87;
      '0'..'9':Dig:=Ord(S[i])-48;
    ELSE Dig:=99;
    END;
    IF Dig>=Base THEN Exit;
    L:=Base*L+Dig;
  END;
  Eval:=TRUE;
END;

{---------------------------------------------------------------------------}
{                                                                           }
{   A p p e n d C h a r                                                     }
{ - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - }
{                                                                           }
{   Hngt das Zeichen C an den String S unter minimalem Stack-Bedarf an.    }
{                                                                           }
{---------------------------------------------------------------------------}
PROCEDURE AppendChar(VAR S:STRING; C:CHAR); ASSEMBLER;
ASM
                les     di,S
                mov     bl,es:[di]
                add     bl,1
                jz      @@Skip
                mov     bh,0
                mov     al,C
                mov     es:[di+bx],al
                mov     es:[di],bl
@@Skip:
END;

{---------------------------------------------------------------------------}
{                                                                           }
{   M a t c h e s                                                           }
{ - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - }
{                                                                           }
{   Prft, ob STR auf das Muster PATTERN (enthlt *, ? (*, und [..]) *)pat }
{   Gro und Kleinschreibung wird unterschieden. Quote-Zeichen ~.           }
{                                                                           }
{---------------------------------------------------------------------------}
FUNCTION Matches(Pattern:STRING; Str:STRING):BOOLEAN;
VAR IdxP,IdxS:INTEGER;
    o:BOOLEAN;
BEGIN
  Matches:=FALSE;
  Str:=Str+#0;
  Pattern:=Pattern+#0;
  IdxP:=1;
  IdxS:=1;
  REPEAT
    CASE Pattern[IdxP] OF
      #0:BEGIN
        Matches:=Str[IdxS]=#0;
        Exit;
      END;
      '?':IF Str[IdxS]=#0 THEN Exit ELSE Inc(IdxP);
      '~':BEGIN
        Inc(IdxP);
        IF (Pattern[IdxP]=#0) OR (Str[IdxS]<>Pattern[IdxP]) THEN Exit;
        Inc(IdxS);
      END;
      '*':BEGIN
        Inc(IdxP);
        IF Pattern[IdxP]=#0 THEN BEGIN
          Matches:=TRUE;
          Exit;
        END;
        REPEAT
         { IF Str[IdxS]=Pattern[IdxP] THEN} BEGIN
            o:=Matches(Copy(Pattern,IdxP,255),Copy(Str,IdxS,255));
            IF o THEN BEGIN
              Matches:=TRUE;
              Exit;
            END;
          END;
          Inc(IdxS);
        UNTIL Str[IdxS]=#0;
        Exit;
      END;
(*    '[':BEGIN
        Inc(IdxP);
        o:=TRUE;
        CASE Pattern[IdxP] OF
          #0:Exit;
          '^':BEGIN
            o:=FALSE;
            Inc(IdxP);
            IF Pattern[IdxP]=#0 THEN Exit;
          END;
        END;
      END; *)
    ELSE
      IF (Pattern[IdxP]=#0) OR (Str[IdxS]<>Pattern[IdxP]) THEN Exit;
      Inc(IdxS);
    END;
    Inc(IdxP);
  UNTIL FALSE;
END;

PROCEDURE SetLength(VAR S:STRING; Len:BYTE);
BEGIN
  WHILE Length(S)<Len DO AppendChar(S,' ');
  IF Length(S)>Len THEN S[0]:=Chr(Len);
END;

FUNCTION StartsWith(S:STRING; SubStr:STRING):BOOLEAN; ASSEMBLER;
ASM
        push    ds
        lds     si,SubStr
        les     di,S
        cld
        lodsb
        cmp     al,es:[di]
        ja      @@False
        inc     di
        mov     cl,al
        mov     ch,0
        repe cmpsb
        jne     @@False
        mov     al,1
        jmp     @@ok
@@False:mov     al,0
@@ok:   pop     ds
END;

END.